# -*- coding: utf-8 -*-
# This file is part of simple-koji-ci project.
#
# simple_koji_ci is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# simple_koji_ci is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with simple_koji_ci.  If not, see <http://www.gnu.org/licenses/>.
"""
Fedmsg consumers that listen to `pagure <http://src.fedoraproject.org>`_.

Authors:    Ralph Bean <rbean@redhat.com>
            Pierre-Yves Chibon <pingou@pingoured.fr>

"""

import logging
import uuid

from requests.packages.urllib3.util import retry
import fedmsg
import fedmsg.consumers
import requests

from simple_koji_ci import __version__
import simple_koji_ci.buildsys
import simple_koji_ci.cache


_log = logging.getLogger(__name__)


def _koji_link(msg):
    instance = msg['msg'].get('instance', 'primary')
    if instance == 'primary':
        base = "http://koji.fedoraproject.org/koji"
    elif instance == 'ppc':
        base = "http://ppc.koji.fedoraproject.org/koji"
    elif instance == 's390':
        base = "http://s390.koji.fedoraproject.org/koji"
    elif instance == 'arm':
        base = "http://arm.koji.fedoraproject.org/koji"
    else:
        raise NotImplementedError("Unhandled instance")

    # One last little switch-a-roo for stg
    if '.stg.' in msg['topic']:
        base = "http://koji.stg.fedoraproject.org/koji"

    return base + "/taskinfo?taskID=%i" \
        % (msg['msg']['id'])


class SimpleKojiCi(fedmsg.consumers.FedmsgConsumer):
    """
    A fedmsg consumer that is the heart of simple-koji-ci.

    This consumer subscribes to the following topics:

    * 'org.fedoraproject.prod.pagure.pull-request.new'
      handled by :method:`SimpleKojiCi.handle_buildsys_scratch`

    * 'org.fedoraproject.prod.pagure.pull-request.comment.added'
      handled by :method:`SimpleKojiCi.handle_anitya_version_update`

    * 'org.fedoraproject.prod.buildsys.task.state.change'
      handled by :method:`SimpleKojiCi.handle_buildsys_scratch`

    """

    # We can do multiple topics like this as of moksha.hub-1.4.4
    # https://github.com/mokshaproject/moksha/pull/25
    topic = [
        # New pull-request coming in
        'org.fedoraproject.stg.pagure.pull-request.new',
        'org.fedoraproject.prod.pagure.pull-request.new',

        # Updates to an existing pull-request
        'org.fedoraproject.stg.pagure.pull-request.comment.added',
        'org.fedoraproject.prod.pagure.pull-request.comment.added',

        # Anyways, we also listen for koji scratch builds to circle back:
        'org.fedoraproject.stg.buildsys.task.state.change',
        'org.fedoraproject.prod.buildsys.task.state.change',
    ]

    config_key = 'simple-koji-ci.enabled'

    def __init__(self, hub):

        # If we're in development mode, rewrite some of our topics so that
        # local playback with fedmsg-dg-replay works as expected.
        if hub.config['environment'] == 'dev':
            # Keep the original set, but append a duplicate set for local work
            prefix, env = hub.config['topic_prefix'], hub.config['environment']
            self.topic = self.topic + [
                '.'.join([prefix, env] + topic.split('.')[3:])
                for topic in self.topic
            ]

        super(SimpleKojiCi, self).__init__(hub)

        if not self._initialized:
            return

        # This is just convenient.
        self.config = self.hub.config

        self.buildsys = simple_koji_ci.buildsys.Koji(
            consumer=self, config=self.config['simple-koji-ci.koji'])

        # Also, set up our global cache object.
        _log.info("Configuring cache.")
        with simple_koji_ci.cache.cache_lock:
            if not hasattr(simple_koji_ci.cache.cache, 'backend'):
                simple_koji_ci.cache.cache.configure(
                    **self.config['simple-koji-ci.cache'])

        # Retrieve the requests configuration; by default requests time out
        # after 15 seconds and are retried up to 3 times.
        self.requests_session = requests.Session()
        self.timeout = (
            self.config.get('simple-koji-ci.connect_timeout', 15),
            self.config.get('simple-koji-ci.read_timeout', 15),
        )
        retries = self.config.get('simple-koji-ci.requests_retries', 3)
        retry_conf = retry.Retry(
            total=retries, connect=retries, read=retries, backoff_factor=1)
        retry_conf.BACKOFF_MAX = 5
        self.requests_session.mount(
            'http://', requests.adapters.HTTPAdapter(max_retries=retry_conf))
        self.requests_session.mount(
            'https://', requests.adapters.HTTPAdapter(max_retries=retry_conf))
        _log.info('Requests timeouts are {}s (connect) and {}s (read)'
                  ' with {} retries'.format(
                    self.timeout[0], self.timeout[1], retries))

        # Build a little store where we'll keep track of what koji scratch
        # builds we have kicked off.  We'll look later for messages indicating
        # that they have completed.
        self.scratch_builds = {}

        _log.info("Simple koji CI is all initialized")

    def consume(self, msg):
        """
        Called when a message arrives on the fedmsg bus.
        """
        topic, msg = msg['topic'], msg['body']
        _log.debug("Received %r" % msg.get('msg_id', None))

        if topic.endswith('pagure.pull-request.new'):
            self.handle_pagure_new_pr(msg)
        elif topic.endswith('buildsys.task.state.change'):
            self.handle_buildsys_scratch(msg)
        else:
            _log.debug("Dropping %r %r" % (topic, msg['msg_id']))
            pass

    def handle_pagure_new_pr(self, msg):
        """
        Message handler for new pull-request opened in pagure.

        Topic: ``org.fedoraproject.pagure.pull-request.new``

        """
        _log.info("Handling pagure msg %r" % msg.get('msg_id', None))

        prid = msg['msg']['pullrequest']['id']
        package = msg['msg']['pullrequest']['project']['name']
        namespace = msg['msg']['pullrequest']['project']['namespace']
        branch_to = msg['msg']['pullrequest']['branch']
        commit = msg['msg']['pullrequest'].get('commit_stop')
        commit_label = ' for %s' % commit[:8] if commit else ''

        try:
            _log.info('Kicking a scratch build for ' + str(package))
            # Kick off a scratch build..
            task_id = self.buildsys.handle(
                package, prid, branch_to=branch_to)

            uid = uuid.uuid4().hex

            # Map that koji task_id to the pull-request we want to flag
            self.scratch_builds[task_id] = dict(
                prid=prid,
                name=package,
                namespace=namespace,
                uid=uid,
                commit=commit,
            )

            # Flag the PR that the build is in progress
            koji_url = self.buildsys.url_for(task_id)
            data = {
                'username': 'simple-koji-ci',
                'status': 'pending',
                'comment': 'Build in progress%s' % commit_label,
                'url': koji_url,
            }

            self._flag_pr(task_id, koji_url, data)

        except Exception as e:
            _log.exception(e)

    def handle_pagure_pr_update(self, msg):
        """
        Message handler for new pull-request opened in pagure.

        Topic: ``org.fedoraproject.pagure.pull-request.comment.added``

        """
        _log.info("Handling pagure msg %r" % msg.get('msg_id', None))

        prid = msg['msg']['pullrequest']['id']
        package = msg['msg']['pullrequest']['project']['name']
        namespace = msg['msg']['pullrequest']['project']['namespace']
        branch_to = msg['msg']['pullrequest']['branch']
        commit = msg['msg']['pullrequest'].get('commit_stop')
        commit_label = ' for %s' % commit[:8] if commit else ''

        if msg['msg']['pullrequest']['status'] != 'Open':
            _log.info('Pull-request not open, discarding')
            return

        if not msg['msg']['pullrequest']['comments']:
            _log.info('This is most odd, we\'re not seeing comments')
            return

        last_comment = msg['msg']['pullrequest']['comments'][-1]

        if not last_comment or last_comment['notification'] is False:
            _log.info('Comment was not a notification, discarding')
            return

        try:
            _log.info('Re-Kicking a scratch build for ' + str(package))
            # Kick off a scratch build..
            task_id = self.buildsys.handle(
                package, prid, branch_to=branch_to)

            uid = uuid.uuid4().hex

            # Map that koji task_id to the pull-request we want to flag
            self.scratch_builds[task_id] = dict(
                prid=prid,
                name=package,
                namespace=namespace,
                uid=uid,
                commit=commit,
            )

            # Flag the PR that the build is in progress
            koji_url = self.buildsys.url_for(task_id)
            data = {
                'username': 'simple-koji-ci',
                'status': 'pending',
                'comment': 'New build in progress%s' % commit_label,
                'url': koji_url,
            }

            self._flag_pr(task_id, koji_url, data)

        except Exception as e:
            _log.exception(e)

    def handle_buildsys_scratch(self, msg):
        """
        Message handler for scratch builds in the build system.

        This handler is used to check up on scratch builds this consumer
        dispatched. When a message arrives it is checked to see if it is
        in a completed state and if it belongs to this consumer. A follow-up
        comment is left on bugs filed when builds are completed.

        Topic: 'org.fedoraproject.prod.buildsys.task.state.change'
        """
        instance = msg['msg']['instance']

        if instance != 'primary':
            _log.debug("Ignoring secondary arch task...")
            return

        method = msg['msg']['method']

        if method != 'build':
            _log.debug("Ignoring non-build task...")
            return

        task_id = msg['msg']['info']['id']
        if task_id not in self.scratch_builds:
            _log.debug("Ignoring [%s] as it's not one of our %d outstanding "
                       "builds" % (str(task_id), len(self.scratch_builds)))
            return

        _log.info("Handling koji scratch msg %r" % msg.get('msg_id'))

        # see koji.TASK_STATES for all values
        done_states = {
            'CLOSED': 'completed',
            'FAILED': 'failed',
            'CANCELED': 'canceled',
        }
        state = msg['msg']['new']
        if state not in done_states:
            return

        koji_url = _koji_link(msg)
        commit = self.scratch_builds[task_id].get('commit')
        commit_label = ' for %s' % commit[:8] if commit else ''
        data = {
            'username': 'simple-koji-ci',
            'status': 'success' if state == 'CLOSED' else 'failure',
            'comment': 'Build %s%s' % (done_states[state], commit_label),
            'url': koji_url,
        }

        self._flag_pr(task_id, koji_url, data)

        del(self.scratch_builds[task_id])

    def _flag_pr(self, task_id, koji_url, data):
        """ Flag a specified pull-request with the given dataload.
        """
        pk_info = self.scratch_builds[task_id]

        pagure_url = self.config["simple-koji-ci.pagure_url"].rstrip('/')
        target_url = '/'.join([
            'api',
            '0',
            pk_info['namespace'],
            pk_info['name'],
            'pull-request',
            str(pk_info['prid']),
            'flag'
        ])
        pagure_url = pagure_url + '/' + target_url
        _log.info('Flagging PR at: %s' % pagure_url)

        headers = {
            "Authorization": "token "
            + self.config["simple-koji-ci.pagure_token"],
            'User-Agent': 'simple-koji-ci %s' % __version__,
        }

        data['uid'] = pk_info['uid']

        _log.debug('payload: %s' % data)

        req = self.requests_session.request(
            method='POST',
            url=pagure_url,
            headers=headers,
            data=data,
        )
        _log.info('Request to %s returned: %s' % (pagure_url, req.status_code))
        _log.debug(req.text)
        if not req.ok:
            _log.debug(req.text)
        else:
            _log.debug('All clear')
